CREATE EXTENSION hll VERSION '2.10';
ALTER EXTENSION hll UPDATE TO '2.11';
ALTER EXTENSION hll UPDATE TO '2.12';
ALTER EXTENSION hll UPDATE TO '2.13';
ALTER EXTENSION hll UPDATE TO '2.14';
ALTER EXTENSION hll UPDATE TO '2.15';
